/*
 * frag_calc.c
 * 2015/7/25 ymaslenko $
 * make a fragment structure (cos(tau), sin(tau), fragment_distance
 * and calculate values
 */

#include "frag_calc.h"
#include "frag_calc_i.h"
#include "e_malloc.h"
#include "bad_angle.h"
#include "dist.h"
#include "coord_i.h"
#include <math.h>

/* ---------------- fc_tau  -----------------------
 * This function sets sin(tau) and cos(tau)
 * from the right sin(tau)_array and cos(tau)_array positions 
 * (equal to dist_array position)
 *  used in fc_get_fragment function
 */
static void fc_tau(const struct frag_calc *frag_calc_t, int idx, float *sin, float *cos) {
    if (idx < 0 || idx >= frag_calc_t->struct_size - 1) {
        *sin = *cos = BAD_ANGLE_FLOAT;
    } else {
        *sin = frag_calc_t->tau_sin[idx + 2];
        *cos = frag_calc_t->tau_cos[idx + 2];
    }
}

/* ---------------- fc_dist  -----------------------
 * This function returns fragment distance
 * from the the dist_array on pos idx
 */
static float fc_dist(const struct frag_calc *frag_calc_t, int idx) {
    if (idx < 0
        || idx >= frag_calc_t->frag_count
        || frag_calc_t->dist[idx] == BAD_DISTANCE_FLOAT)
        return BAD_DISTANCE_FLOAT;

    return frag_calc_t->dist[idx];
}


/* ---------------- fc_init  -----------------------
 * This function initializes the frag_calc structure
 */
struct frag_calc *fc_init(struct coord *structure, size_t frag_len) {
    struct frag_calc *frag_calc_t = E_CALLOC(1, sizeof(struct frag_calc));
    size_t i;
    frag_calc_t->frag_count_healthy = structure->size - frag_len + 1; /*'healthy' fragments*/
    frag_calc_t->frag_count = frag_calc_t->frag_count_healthy; /*NO BROCKEN FRAGS*/
    frag_calc_t->frag_len = frag_len;
    frag_calc_t->struct_size = structure->size;

    coord_calc_tau(structure); /*tau_array for the complette protein structure with bad_angles
                                  on pos tau[0], tau[1] and tau[last]. use fc_tau to make array_pos 
                                  equal to dist_array*/
    frag_calc_t->tau_sin = E_CALLOC(structure->size, sizeof(*frag_calc_t->tau_sin));
    frag_calc_t->tau_cos = E_CALLOC(structure->size, sizeof(*frag_calc_t->tau_cos));

    for (i = 0; i < structure->size; i++) { /* use fc_tau to make array_pos equal to dist_array*/
        float tau = structure->tau[i];
        if (tau != BAD_ANGLE_FLOAT) {
            frag_calc_t->tau_sin[i] = sinf(tau);
            frag_calc_t->tau_cos[i] = cosf(tau);
        } else {
            frag_calc_t->tau_sin[i] = BAD_ANGLE_FLOAT;
            frag_calc_t->tau_cos[i] = BAD_ANGLE_FLOAT;
        }
    }

    frag_calc_t->dist = E_CALLOC(frag_calc_t->frag_count_healthy, sizeof(*frag_calc_t->dist));

    for (i = 0; i < frag_calc_t->frag_count_healthy; i++) {
        frag_calc_t->dist[i] = dist_frag_coord(structure, i, frag_len, 0);
    }

    return frag_calc_t;
}


/* ---------------- fc_get_fragment  -----------------------
 * This function returns the fragment_array
 * cos(tau) and sin(tau) are on the positions from fragment[0] to fragment[last_pos-1] 
 * and fragment_distance is fragment[last_pos]
 */
float *fc_get_fragment(const struct frag_calc *frag_calc_t, size_t position) {

    size_t tau_count = frag_calc_t->frag_len - 3;    /*number of tau_angles in a fragment*/
    size_t fragment_size = tau_count * 2 + 1;        /*dimension or lenght of the fragment_array*/
    float *fragment = E_CALLOC(fragment_size, sizeof(float));

    int atom_start = position;                       /*NO BROCKEN FRAGS*/
    fragment[0] = fc_dist(frag_calc_t, atom_start);  /*fragment_distance on the first position*/
    size_t i;

    for (i = 0; i < tau_count; i++) {
        float sin, cos;
        fc_tau(frag_calc_t, atom_start + i, &sin, &cos); /* makes tau_array_pos equal to dist_array*/
        fragment[2 * i + 1] = cos;
        fragment[2 * i + 2] = sin;
    }
    return fragment;
}

/* ---------------- fc_get_fragment_count_healthy  -----------------------
 * This function returns number of healthy fragments in the protein structure
 */
size_t fc_get_fragment_count_healthy(const struct frag_calc *frag_calc_t) {
    return frag_calc_t->frag_count_healthy;
}

/* ---------------- fc_get_fragment_count  -----------------------
 * This function returns number of fragments in the protein structure
 * healthy_fragments
 */
size_t fc_get_fragment_count(const struct frag_calc *frag_calc_t) {
    return frag_calc_t->frag_count;
}


/* ---------------- fc_destroy  -----------------------
 * This function destroys tau_sin, tau_cos and dist arrays
 */
void fc_destroy(struct frag_calc *frag_calc_t) {
    if (frag_calc_t) {
        free_if_not_null(frag_calc_t->tau_sin);
        free_if_not_null(frag_calc_t->tau_cos);
        free_if_not_null(frag_calc_t->dist);
        free(frag_calc_t);
    }
}