/*
 * Read up a classification based on structure
 * $Id: read_ac_tau.c,v 1.2015/09/10
 * modified ymaslenko  2015/7/25 Exp $
 */

#ifndef _XOPEN_SOURCE
#   define _XOPEN_SOURCE 500
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "classifyStructure.h"
#include "class_model_tau.h"
//#include "class_model.h"
#include "const_clss7.h"
#include "coord.h"
#include "coord_i.h"
#include "e_malloc.h"
#include "matrix.h"
#include "mprintf.h"
#include "prob_vec.h"
#include "prob_vec_i.h"
#include "read_ac.h"
#include "read_ac_i.h"
#include "read_tau_strct.h"
#include "read_tau_strct_i.h"
#include "read_seq_i.h"
#include "seq.h"
//#include "seq_i.h"
#include "seqprof.h"
#include "yesno.h"
#include "frag_calc_i.h"
#include "frag_calc.h"
#include "amino_a.h"


/* 24 MAR 2006 TS */
/* ---------------- tau_strct_size -----------------------------
 * Return the fragment length size associated with a
 * classification. This is *not* the number of classes. It will
 * typically be a number between 4 and 9.
 */
size_t
tau_strct_size(const struct tau_strct_clssfcn *clssfcn) {
    return clssfcn->n_att;
}

/* ---------------- tau_strct_nclass ---------------------------
 * Return the number of classes in a classification.
 */
size_t
tau_strct_nclass(const struct tau_strct_clssfcn *clssfcn) {
    return clssfcn->strct->n_class;
}


/* ---------------- tau_strct_clssfcn_destroy ------------------
 * This clean up function will be called by the perl interface.
 */
void
tau_strct_clssfcn_destroy(struct tau_strct_clssfcn *clssfcn) {
    /*not to free the clssfcn if it is a constant structure from i.e tau_strct_clss6.c
      included from const_clss.h. It is here as perl's garbage collector calls destroy
      by itself.*/
    if (!clssfcn || (clssfcn == pstrct_clss7()))
        return;
    if (clssfcn->strct)
        clssfcn_destroy(clssfcn->strct);
    free(clssfcn);
}

/* ---------------- tau_strct_clssfcn_read ---------------------
 * Given a pointer to a filename, read in the data.
 * This is an interface function, visible to the outside world.
 */
struct tau_strct_clssfcn *
tau_strct_clssfcn_read(const char *fname, const float abs_error_dist, const float abs_error_tau) {
//    struct clssfcn7 *tau_class;
    struct tau_strct_clssfcn *clssfcn;

    clssfcn = E_MALLOC (sizeof(struct tau_strct_clssfcn));
    clssfcn->strct = NULL;
    /* Read the structure information */

    if ((clssfcn->strct = tau_get_clssfcn(fname, abs_error_dist, abs_error_tau)) == NULL) {
        return NULL;
    }

    /* Read the residue information */
    clssfcn->n_att = clssfcn->strct->dim * 0.5 + 1;

    return clssfcn;
}

/* ---------------- tau_strct_2_prob_vec_calc  -----------------------
 * This calculates probabilities using combinations of
 * sequence and structure, sequence profile and structure,
 * sequence only, sequence profile only, and structure only.
 */
static struct prob_vec *
tau_strct_2_prob_vec_calc(struct coord *structure,
                          const size_t size,
                          const struct tau_strct_clssfcn *cmodel,
                          const enum yes_no norm) {

    size_t i, j, compnd_len;
    struct prob_vec *pvec;
    float **tau_prob;
    struct clssfcn7 *tau_class;
    const char *this_sub = "tau_strct_2_prob_vec";

    size_t frag_len = (size_t)((cmodel->strct->dim - 1) * 0.5 + 3); /*fragment length*/
    const size_t tau_count = frag_len - 3;                          /*number of tau_angles in a fragment*/
    const size_t frag_num = size - frag_len + 1;                    /*classical fragments*/

    const size_t n_pvec = frag_num;

    if ((pvec = new_pvec(frag_len, size, n_pvec,
                         cmodel->strct->n_class)) != NULL) {
        pvec->n_pvec = n_pvec;
        for (i = 0; i < n_pvec; i++)   /* Initialize mship */
            for (j = 0; j < cmodel->strct->n_class; j++)
                pvec->mship[i][j] = cmodel->strct->class_weight7[j];

        if (structure) {
            struct frag_calc *fc = fc_init(structure, frag_len);

/*#pragma omp parallel for*/
            for (i = 0; i < n_pvec; i++) {
                float *fragment = fc_get_fragment(fc, i);

                if (tau_computeMembership(pvec->mship[i], fragment, cmodel->strct) == NULL) {
                    size_t j;
                    for (j = 0; j < cmodel->strct->n_class; j++) {
                        if (pvec->mship[i][j] == 0) {
                            mprintf("there is a zero value at %ld/%ld\n", i + 1, j);
                        }
                    }
                    prob_vec_destroy(pvec);
                    err_printf(this_sub, "No Prob Vec calculated!\n");
                }

                free_if_not_null(fragment);
                fragment = NULL;
            }
            fc_destroy(fc);
        }

        if (norm == YES) {
            for (i = 0; i < pvec->n_pvec; i++) {
                double sum = 0.0;
                for (j = 0; j < cmodel->strct->n_class; j++)
                    sum += pvec->mship[i][j];
                for (j = 0; j < cmodel->strct->n_class; j++)
                    pvec->mship[i][j] /= sum;
            }
            pvec->norm_type = PVEC_TRUE_PROB;
        }
    }

    if (structure)
        compnd_len = structure->compnd_len;
    else
        compnd_len = 0;

    if (compnd_len > 0) {           /* finally read compound*/
        pvec->compnd = E_MALLOC(compnd_len * sizeof(char));
        memmove(pvec->compnd, structure->compnd, compnd_len);
    }
    pvec->compnd_len = compnd_len;
    return pvec;
}

/* ---------------- tau_strct_2_prob_vec  -------------------------
 * This calculates probabilities, but using only structure terms.
 */
struct prob_vec *
tau_strct_2_prob_vec(struct coord *structure,
                     const struct tau_strct_clssfcn *cmodel, const int norm) {
    const char *this_sub = "strct_2_prob_vec_TAU";

    if (!structure) {
        err_printf(this_sub, "No Structure Input!\n");
        return NULL;
    }
    if (!cmodel || !cmodel->strct) {
        err_printf(this_sub, "No Classification Input!\n");
        return NULL;
    }

    return tau_strct_2_prob_vec_calc(structure, structure->size,
                                     cmodel, norm);
}

